# Clapper Enhancer yt-dlp
# Copyright (C) 2024 Rafał Dzięgiel <rafostar.github@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import gi
gi.require_version('GLib', '2.0')
gi.require_version('GObject', '2.0')
gi.require_version('Gio', '2.0')
gi.require_version('Gst', '1.0')
gi.require_version('Clapper', '0.0')
from gi.repository import GLib, GObject, Gio, Gst, Clapper

try:
    from yt_dlp import YoutubeDL
except ImportError:
    YoutubeDL = None

if YoutubeDL:
    from yt_dlp.extractor import gen_extractor_classes
    from clapper_yt_dlp_overrides import BLACKLIST, ClapperYoutubeIE

import clapper_yt_dlp_dash as dash
import clapper_yt_dlp_hls as hls
import clapper_yt_dlp_direct as direct

YTDL_OPTS = {
    'quiet': True,
    'color': 'never', # no color in exceptions
    'ignoreconfig': True,
    'extract_flat': 'in_playlist',
    'extractor_args': {
        'youtube': {
            'skip': ['translated_subs'],
            'player_client': ['ios']
        }
    }
}

class ClapperYtDlp(GObject.Object, Clapper.Extractable):
    _ytdl = None

    def __init__(self):
        if YoutubeDL:
            self._ytdl = YoutubeDL(YTDL_OPTS, auto_init=False)

            self._ytdl.add_info_extractor(ClapperYoutubeIE())

            for ie in gen_extractor_classes():
                if ie.ie_key() not in BLACKLIST:
                    self._ytdl.add_info_extractor(ie)

    def do_extract(self, uri: GLib.Uri, harvest: Clapper.Harvest, cancellable: Gio.Cancellable):
        if not YoutubeDL:
            raise GLib.Error('Could not import "yt-dlp". Please check your installation.')

        # Not used during init, so we can alter it here
        self._ytdl.params['noplaylist'] = True

        try:
            info = self._ytdl.extract_info(uri.to_string(), download=False)
        except Exception as e:
            raise GLib.Error(str(e))

        # Check if cancelled during extraction
        if cancellable.is_cancelled():
            return False

        if (manifest := dash.generate_manifest(info)):
            media_type = 'application/dash+xml'
        elif (manifest := hls.generate_manifest(info)):
            media_type = 'application/x-hls'
        elif (manifest := direct.generate_manifest(info)):
            media_type = 'text/uri-list'
        else:
            raise GLib.Error('Could not generate playable manifest')

        # Check if cancelled during manifest generation
        if cancellable.is_cancelled():
            return False

        harvest.fill_with_text(media_type, manifest)

        if (val := info.get('title')):
            harvest.tags_add(Gst.TAG_TITLE, val)
        if (val := info.get('duration')):
            value = GObject.Value()
            value.init(GObject.TYPE_UINT64)
            value.set_uint64(val * Gst.SECOND)
            harvest.tags_add(Gst.TAG_DURATION, value)
        if (val := info.get('chapters')):
            for index, chap in enumerate(val):
                title, start, end = chap.get('title'), chap.get('start_time'), chap.get('end_time')
                harvest.toc_add(Gst.TocEntryType.CHAPTER, title, start, end)

        # XXX: We just take headers from any format here, do we need to find/combine some?
        for fmt in info['formats']:
            if (hdrs := fmt.get('http_headers')):
                [harvest.headers_set(key, val) for key, val in hdrs.items()]
                break

        return True



# Clapper Enhancer yt-dlp
# Copyright (C) 2024 Rafał Dzięgiel <rafostar.github@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import io

def _make_manifest(info, vext):
    best_format = None

    for fmt in info['formats']:
        if (
                fmt.get('protocol') != 'https'
                or fmt.get('video_ext') != vext
                or not fmt.get('url')
        ):
            continue

        if not best_format:
            best_format = fmt
            continue

        if vext != 'none':
            height = fmt.get('height') or 0
            best_height = best_format.get('height') or 0

            fps = fmt.get('fps') or 0
            best_fps = best_format.get('fps') or 0

            if (
                    height > best_height
                    or (height == best_height and fps > best_fps)
            ):
                best_format = fmt
                continue

        tbr = fmt.get('tbr') or 0
        best_tbr = best_format.get('tbr') or 0

        if (tbr > best_tbr):
            best_format = fmt

    if not best_format:
        return None

    manifest = io.StringIO()
    manifest.write(best_format['url'])

    return manifest

def generate_manifest(info):
    if (
            (manifest := _make_manifest(info, 'mp4'))
            or (manifest := _make_manifest(info, 'none')) # Audio only
    ):
        return manifest.getvalue()

    return None



# Clapper Enhancer yt-dlp
# Copyright (C) 2024 Rafał Dzięgiel <rafostar.github@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import xml.etree.ElementTree as ET
from math import gcd

def _get_aspect_ratio(width, height):
    div = gcd(width, height)
    return f'{width // div}:{height // div}'

def _insert_representations(adapt, formats):
    for fmt in formats:
        rep = ET.SubElement(adapt, 'Representation')
        rep.set('id', fmt['format_id'])
        rep.set('bandwidth', str(int(fmt['tbr'] * 1000))) # bps

        vcodec = None if fmt['vcodec'] == 'none' else fmt['vcodec']
        acodec = None if fmt['acodec'] == 'none' else fmt['acodec']

        if vcodec and acodec:
            rep.set('codecs', f'{vcodec},{acodec}')
        elif not acodec:
            rep.set('codecs', vcodec)
        elif not vcodec:
            rep.set('codecs', acodec)

        if vcodec:
            if (width := fmt.get('width') or 0) > 0:
                rep.set('width', str(width))
            if (height := fmt.get('height') or 0) > 0:
                rep.set('height', str(height))
                if width > 0:
                    rep.set('sar', _get_aspect_ratio(width, height))
            if (fps := fmt.get('fps') or 0) > 0:
                rep.set('frameRate', str(fps))

        if acodec and (aud_ch := fmt.get('audio_channels') or 0) > 0:
            ch_conf = ET.SubElement(rep, 'AudioChannelConfiguration')
            ch_conf.set('schemeIdUri', 'urn:mpeg:dash:23003:3:audio_channel_configuration:2011')
            ch_conf.set('value', str(aud_ch))

        base_url = ET.SubElement(rep, 'BaseURL')
        base_url.text = fmt['url']

        # Existence ensured before calling this function
        streaming = fmt['streaming_options']
        init_range = streaming['init_range']
        index_range = streaming['index_range']

        initialization = ET.SubElement(rep, 'Initialization')
        initialization.set('range', init_range)

        segment_base = ET.SubElement(rep, 'SegmentBase')
        segment_base.set('indexRange', index_range)
        segment_base.set('indexRangeExact', 'true')

def _add_adaptation_set(period, info, vcoding, acoding, lang=None):
    max_w = 0
    max_h = 0
    max_fps = 0
    ext = None
    formats = []

    for fmt in info['formats']:
        # Ensure all required values before creating an adaptation set
        if (
                (lang and fmt.get('language') != lang)
                or not ((val := fmt.get('container')) and val.endswith('_dash'))
                or not fmt.get('vcodec', 'none').startswith(vcoding)
                or not fmt.get('acodec', 'none').startswith(acoding)
                or not ((val := fmt.get('ext')) and not val == 'none')
                or not ((val := fmt.get('format_id')) and not val.endswith('-drc'))
                or not (val := fmt.get('tbr') or 0) > 0
        ):
            continue

        if (url := fmt.get('url')): # isoff-on-demand
            if (
                    not (streaming := fmt.get('streaming_options'))
                    or not streaming.get('init_range')
                    or not streaming.get('index_range')
            ):
                continue
        else: # FIXME: isoff-live
            continue

        if vcoding != 'none':
            # Ignore ultralow video qualities
            if (height := fmt.get('height') or 0) < 240:
                continue

            width = fmt.get('width') or 0
            fps = fmt.get('fps') or 0

            max_w = max(max_w, width)
            max_h = max(max_h, height)
            max_fps = max(max_fps, fps)

        if acoding != 'none':
            # Ignore ultralow audio qualities
            if (val := fmt.get('format_note')) == 'ultralow':
                continue

        # All is good
        formats.append(fmt)

    if len(formats) == 0:
        return False

    # yt-dlp uses 'm4a' while DASH expects 'mp4'
    ext = formats[0]['ext']
    ext = {'m4a': 'mp4'}.get(ext, ext)

    cnt_type = 'video' if vcoding != 'none' else 'audio'

    adapt = ET.SubElement(period, 'AdaptationSet')
    adapt.set('contentType', cnt_type)
    adapt.set('mimeType', f'{cnt_type}/{ext}')
    adapt.set('subsegmentAlignment', 'true')
    adapt.set('subsegmentStartsWithSAP', '1')

    if max_w > 0:
        adapt.set('maxWidth', str(max_w))
    if max_h > 0:
        adapt.set('maxHeight', str(max_h))
    if max_w > 0 and max_h > 0:
        div = gcd(max_w, max_h)
        adapt.set('par', f'{max_w // div}:{max_h // div}')
    if max_fps > 0:
        adapt.set('maxFrameRate', str(max_fps))
    if lang:
        # Convert to ISO-639
        adapt.set('lang', lang.split('-')[0])

    _insert_representations(adapt, formats)

    return True

def _add_audio_adaptation_sets(period, info, acodings):
    success = False
    languages = []

    # Find languages and determine default one
    preference = 0
    default_lang = None
    for fmt in info['formats']:
        if not (lang := fmt.get('language')):
            continue

        if (lang_pref := fmt.get('language_preference') or 0) > preference:
            # Append remembered language if not in array yet before replacement
            if default_lang and default_lang not in languages:
                languages.append(default_lang)

            # Replace with new default
            default_lang = lang
            preference = lang_pref
        elif lang != default_lang and lang not in languages:
            languages.append(lang)

    if default_lang:
        languages.insert(0, default_lang)

    for acoding in acodings:
        # Add an adaptiation set for each language
        for lang in languages:
            success |= _add_adaptation_set(period, info, 'none', acoding, lang)

        # Fallback for undetermined language
        if not success:
            success |= _add_adaptation_set(period, info, 'none', acoding)

        # At least one language for acoding added
        if success:
            break

    return success

def generate_manifest(info):
    success = False

    if (duration := int(info.get('duration') or 0)):
        ns_xsi = 'http://www.w3.org/2001/XMLSchema-instance'

        ET.register_namespace('xsi', ns_xsi)
        qname = ET.QName(ns_xsi, 'schemaLocation')

        dur_pts = f'PT{duration}S'
        buf_pts = f'PT{min(2, duration)}S'

        mpd = ET.Element('MPD')
        mpd.set(qname, 'urn:mpeg:dash:schema:mpd:2011 DASH-MPD.xsd')
        mpd.set('xmlns', 'urn:mpeg:dash:schema:mpd:2011')
        mpd.set('type', 'static')
        mpd.set('mediaPresentationDuration', dur_pts)
        mpd.set('minBufferTime', buf_pts)
        mpd.set('profiles', 'urn:mpeg:dash:profile:isoff-on-demand:2011')

        period = ET.SubElement(mpd, 'Period')

        vcodings = ['avc1', 'av01', 'hev1', 'vp09']
        acodings = [] # Possibilities depend on selected vcoding

        for vcoding in vcodings:
            if (success := _add_adaptation_set(period, info, vcoding, 'none')):
                if vcoding in ['avc1', 'av01', 'hev1']:
                    acodings += ['mp4a']
                elif vcoding in ['vp09']:
                    acodings += ['opus']

                break

        # No video, allow any supported audio for audio-only
        if not success:
            acodings = ['mp4a', 'opus']

        success |= _add_audio_adaptation_sets(period, info, acodings)

        # If separate failed, try combined
        if not success:
            for vcoding in vcodings:
                for acoding in acodings:
                    if (success := _add_adaptation_set(period, info, vcoding, acoding)):
                        break

    if not success:
        return None

    ET.indent(mpd)

    return ET.tostring(mpd, encoding='unicode', xml_declaration=True)



# Clapper Enhancer yt-dlp
# Copyright (C) 2024 Rafał Dzięgiel <rafostar.github@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import re

from yt_dlp.extractor.youtube import YoutubeIE
from yt_dlp.utils import parse_m3u8_attributes

BLACKLIST = [
    'Youtube',
    'Generic'
]

class ClapperYoutubeIE(YoutubeIE):
    @classmethod
    def ie_key(cls):
        return cls.__name__[7:-2]

    _ITAG_PATTERN = re.compile(r'/itag/(\d+)')

    def _extract_formats_and_subtitles(self, streaming_data, video_id, player_url, live_status, duration):
        *formats, subtitles = super()._extract_formats_and_subtitles(streaming_data, video_id, player_url, live_status, duration)

        for fmt in formats:
            if not (format_id := fmt.get('format_id')):
                continue

            itag = format_id.split('-')[0]
            lang = (language := fmt.get('language')) and language.split('-')[0]

            for resp_data in streaming_data:
                resp_found = False

                for yt_fmt in resp_data.get('adaptiveFormats'):
                    # Find format with the same itag
                    if (
                            itag != str(yt_fmt.get('itag') or 0)
                            or yt_fmt.get('isDrc', False)
                    ):
                        continue

                    # Special cases in YT formats: "en.4"
                    yt_lang = (yt_atrack := yt_fmt.get('audioTrack')) and yt_atrack['id'].split('-')[0].split('.')[0]

                    # Youtube uses the same itag for different languages, so language must also match
                    if (lang != yt_lang):
                        continue

                    init_range = yt_fmt.get('initRange')
                    index_range = yt_fmt.get('indexRange')

                    if not init_range or not index_range:
                        continue

                    init_start = int(init_range.get('start') or 0)
                    init_end = int(init_range.get('end') or 0)

                    index_start = int(index_range.get('start') or 0)
                    index_end = int(index_range.get('end') or 0)

                    if init_end <= init_start or index_end <= index_start:
                        continue

                    fmt['streaming_options'] = {
                        'init_range': f'{init_start}-{init_end}',
                        'index_range': f'{index_start}-{index_end}'
                    }

                    resp_found = True
                    break

                if resp_found:
                    break

        return *formats, subtitles

    def _parse_m3u8_formats_and_subtitles(
            self, m3u8_doc, m3u8_url=None, ext=None, entry_protocol='m3u8_native',
            preference=None, quality=None, m3u8_id=None, live=False, note=None,
            errnote=None, fatal=True, data=None, headers={}, query={},
            video_id=None):
        formats, subtitles = super()._parse_m3u8_formats_and_subtitles(
                m3u8_doc, m3u8_url, ext, entry_protocol,
                preference, quality, m3u8_id, live, note,
                errnote, fatal, data, headers, query,
                video_id)

        for line in m3u8_doc.splitlines():
            if not line.startswith('#EXT-X-STREAM-INF:'):
                continue

            attributes = parse_m3u8_attributes(line)
            bandwidth = attributes.get('BANDWIDTH')
            audio_id = attributes.get('AUDIO')
            captions_id = attributes.get('CLOSED-CAPTIONS')

            if not bandwidth or not (audio_id or captions_id):
                continue

            for fmt in formats:
                # Format ID might not be extracted here yet
                if not (format_id := fmt.get('format_id')):
                    match = self._ITAG_PATTERN.search(fmt['url'])
                    if not (format_id := match and match.group()):
                        continue

                group_id = format_id.split('-')[0]

                # Fix missing audio codecs
                if (group_id == audio_id and fmt.get('acodec', 'none') == 'none'):
                    codecs_arr = attributes.get('CODECS').split(',')

                    if len(codecs_arr) > 1:
                        fmt['acodec'] = codecs_arr[1]

                # Fix unknown audio/captions stream for video
                if int((fmt.get('tbr') or 0) * 1000) == int(bandwidth):
                    fmt['audio_id'] = audio_id
                    fmt['captions_id'] = captions_id

        return formats, subtitles



# Clapper Enhancer yt-dlp
# Copyright (C) 2024 Rafał Dzięgiel <rafostar.github@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import io

def _add_x_media(manifest, fmt):
    acodec = None if fmt['acodec'] == 'none' else fmt['acodec']
    media_type = 'AUDIO' if acodec else 'CLOSED-CAPTIONS'

    manifest.write(f'\n#EXT-X-MEDIA:TYPE={media_type}')

    fmt_id = fmt['format_id'].split('-')[0]
    manifest.write(f',GROUP-ID="{fmt_id}"')

    is_default = 'YES'

    # NOTE: GStreamer expects variants of single language to have the same NAME
    if ((language := fmt.get('language')) and language != 'none'):
        # Convert to ISO-639
        lang = language.split('-')[0]
        manifest.write(f',LANGUAGE="{lang}"')
        manifest.write(f',NAME="{language}"')
        is_default = 'YES' if (fmt.get('language_preference') or 0) > 0 else 'NO'
    else:
        manifest.write(f',NAME="Default"')

    manifest.write(f',DEFAULT={is_default},AUTOSELECT=YES')

    url = fmt['url']
    manifest.write(f',URI="{url}"')

def _add_x_stream_inf(manifest, fmt, audio, captions):
    # Existence ensured before calling this function
    bandwidth = int(fmt['tbr'] * 1000)
    manifest.write(f'\n#EXT-X-STREAM-INF:BANDWIDTH={bandwidth}')

    vcodec = None if fmt['vcodec'] == 'none' else fmt['vcodec']
    acodec = None if fmt['acodec'] == 'none' else fmt['acodec']

    if not acodec and audio:
        acodec = None if audio['acodec'] == 'none' else audio['acodec']

    if vcodec and acodec:
        manifest.write(f',CODECS="{vcodec},{acodec}"')
    else:
        codec = vcodec if vcodec else acodec
        manifest.write(f',CODECS="{codec}"')

    width = fmt.get('width') or 0
    height = fmt.get('height') or 0
    if width > 0 and height > 0:
        manifest.write(f',RESOLUTION={width}x{height}')

    if ((fps := fmt.get('fps') or 0) > 0):
        manifest.write(f',FRAME-RATE={fps}')

    if (dyn_range := fmt.get('dynamic_range')):
        manifest.write(f',VIDEO-RANGE={dyn_range}')

    if vcodec and audio:
        audio_id = fmt['audio_id']
        manifest.write(f',AUDIO="{audio_id}"')

    if vcodec and captions:
        captions_id = fmt['captions_id']
        manifest.write(f',CLOSED-CAPTIONS="{captions_id}"')

    url = fmt['url']
    manifest.write(f'\n{url}')

def _insert_streams(manifest, formats, matches):
    for fmt in formats:
        have_bandwidth = (fmt.get('tbr') != None)

        if have_bandwidth:
            audio = None
            captions = None

            if (audio_id := fmt.get('audio_id')):
                audio = next(filter(lambda m: m['format_id'].split('-')[0] == audio_id, matches), None)
            if (captions_id := fmt.get('captions_id')):
                captions = next(filter(lambda m: m['format_id'].split('-')[0] == captions_id, matches), None)

            _add_x_stream_inf(manifest, fmt, audio, captions)
        else:
            _add_x_media(manifest, fmt)

def _add_streams(manifest, info, vcoding, acoding):
    formats = []
    matches = []

    for fmt in info['formats']:
        # Ensure all required values before creating adding stream
        if (
                not ((val := fmt.get('protocol')) and val == 'm3u8_native')
                or not fmt.get('url')
                or not fmt.get('vcodec', 'none').startswith(vcoding)
                or not fmt.get('acodec', 'none').startswith(acoding)
                or not ((val := fmt.get('format_id')) and not val.endswith('-drc'))
        ):
            continue

        if vcoding != 'none':
            # Ignore ultralow video qualities
            if (height := fmt.get('height') or 0) < 240:
                continue

            if ((audio_id := fmt.get('audio_id')) and not any(m['format_id'].split('-')[0] == audio_id for m in matches)):
                for audio_fmt in info['formats']:
                    if (audio_fmt.get('format_id', '').split('-')[0] == audio_id):
                        matches.append(audio_fmt)

        if acoding != 'none':
            # Ignore ultralow audio qualities
            if (val := fmt.get('format_note')) == 'ultralow':
                continue

        # All is good
        formats.append(fmt)

    if len(formats) == 0:
        return False

    _insert_streams(manifest, formats, matches)

    return True

def _make_manifest(info, vcoding, acoding, separate=False):
    manifest = io.StringIO()
    success = True

    manifest.write('#EXTM3U\n#EXT-X-INDEPENDENT-SEGMENTS')

    if separate:
        # Start with audio stream as video streams often point to audio
        if acoding != 'none':
            success &= _add_streams(manifest, info, 'none', acoding)
        if vcoding != 'none':
            success &= _add_streams(manifest, info, vcoding, 'none')
    else:
        success = _add_streams(manifest, info, vcoding, acoding)

    if not success:
        return None

    return manifest

def generate_manifest(info):
    if (
            (manifest := _make_manifest(info, 'avc1', 'mp4a', True)) # Video + Audio separately
            or (manifest := _make_manifest(info, 'avc1', 'mp4a')) # Video + Audio combined
            or (manifest := _make_manifest(info, 'none', 'mp4a')) # Audio only
    ):
        return manifest.getvalue()

    return None
